# -*- coding: utf-8 -*-
# Project : flacon
# File : converter.py
# (c) Alexander Sokolov asokol@mail.ru
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This file is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

import errno
import exceptions
import re
import sys
import time
import subprocess

from PyQt4 import QtCore 
from PyQt4.QtCore import *

from tracklist import * 
from preferences import *


class AudioFormat:
    UNKNOWN = ''
    WAV     = 'WAV'
    FLAC    = 'FLAC'
    OGG     = 'OGG'
    MP3     = 'MP3'    
    WM      = 'WM'
    APE     = 'APE'    
    
    
class Abort(Exception):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self):
        Exception.__init__(self, '')


class ProgramNotfound(Exception):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, msg):
        Exception.__init__(self, msg)


class Converter(QObject):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, parent = None):
        QObject.__init__(self, parent)
        self._trackList = None;
        self._threads = []
        self._encoderClass = None
        self._queue = []
        self._threadCount = 0


   #****************************************************************
   #
   #****************************************************************
    def start(self, trackList, format):
        self._trackList = trackList;
        self._format = format
        
        if format == AudioFormat.FLAC: self._encoderClass = FlacEncoder
        if format == AudioFormat.OGG:  self._encoderClass = OggEncoder
        if format == AudioFormat.WAV:  self._encoderClass = WavEncoder
        if format == AudioFormat.MP3:  self._encoderClass = Mp3Encoder
        
        if not self._encoderClass:
            self.error("Unknown format.")
            raise Exception("Unknown format.")

        try:
          Splitter.check()  
          self._encoderClass.check()
        except ProgramNotfound,e:
            msg = self.tr("Conversion is not possible.\nThe following programs are not found:\n%1").arg(unicode(e))
            self.error(msg)
            raise Exception(unicode(msg)) 


        (self._threadCount, ok) = qApp.preferences.value('Encoder/ThreadCount').toInt()

        if self._threadCount<1: 
            self._threadCount = max(6, QThread.idealThreadCount())
               
        splitter = Splitter(self._trackList)
        self._threads.append(splitter)
        self.connect(splitter, SIGNAL("fileReady(QString, int)"), self.trackSplitted)
        self.connect(splitter, SIGNAL("error(QString)"), self.error)
        self.connect(splitter, SIGNAL("finished()"), self.finished)
        self.connect(splitter, SIGNAL("progress(int, QString, int)"), SIGNAL("progress(int, QString, int)"))
        self._startedTime = time.time()
        splitter.start()


   #****************************************************************
   # Error handler, terminate splitter and all encoders
   #****************************************************************
    def error(self, msg):
        del self._queue[:]
        for thread in self._threads:
            try:
                thread.stop()
                thread.wait()
            except:
                pass
            
        self.emit(SIGNAL("error(QString)"), msg)
        self.emit(SIGNAL("finished()"))
     

   #****************************************************************
   # 
   #****************************************************************
    def startThread(self):
        if len(self._queue) and (len(self._threads)<self._threadCount):
            (fileName, trackIndex) = self._queue.pop(0)
            track = self._trackList.track(trackIndex)
            encoder = self._encoderClass(fileName, track)
            self._threads.append(encoder)
            self.connect(encoder, SIGNAL("error(QString)"), self.error)
            self.connect(encoder, SIGNAL("finished()"), self.finished)
            self.connect(encoder, SIGNAL("progress(int, QString, int)"), SIGNAL("progress(int, QString, int)"))
            encoder.start()


   #****************************************************************
   # 
   #****************************************************************
    def finished(self):
        try:
            self._threads.remove(self.sender())
        except:
            pass
        
        self.startThread()

        if not len(self._threads):
            self.emit(SIGNAL("finished()"))
            length = time.time() - self._startedTime
            print "Encoding time %d sec." %  length


   #****************************************************************
   # 
   #****************************************************************
    def trackSplitted(self, fileName, trackIndex):
        self.emit(SIGNAL("progress(int, QString, int)"), trackIndex, self.tr('Queued'), -1)
        self._queue.append((fileName, trackIndex))
        self.startThread()


   #****************************************************************
   # 
   #****************************************************************
    def cancel(self):
        for (fileName, trackIndex) in self._queue:
            self.emit(SIGNAL("progress(int, QString, int)"), trackIndex, self.tr('Aborted'), -1)      
            
        del self._queue[:]
        
        for thread in self._threads:
            thread.stop()
            thread.wait()
            
        self.emit(SIGNAL("finished()"))
        
        
   #****************************************************************
   # 
   #****************************************************************
    def isRunning(self):
        return len(self._threads) > 0



class Splitter(QThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, trackList):
        QThread.__init__(self)
        self._trackList = trackList
        self._errors = QStringList()
        self._cancel = False
        self._currentTrack = -1
        self._workDir = None
        
        
   #****************************************************************
   # Create directory for temporary files 
   #****************************************************************
    def createDir(self, workDir):
        if not workDir.mkpath('.'):
            e = self.tr("Can't create directory \"%1\".").arg(workDir.path())
            self.emit(SIGNAL("error(QString)"), e)
            raise Exception(unicode(e))           
    
    
   #****************************************************************
   # Split ape to temporary dir 
   #****************************************************************
    def run(self):
        self._cancel = False
        self._workDir = self._trackList.track(0).outFile().dir()
        self.createDir(self._workDir)
    
        args = []
        args.append(unicode(qApp.preferences.value("Programs/shntool").toString()))
        args.append('split')
        args.append('-w')
        args.append('-O');      args.append('always')
        args.append('-n');      args.append('%04d')
        args.append('-t');      args.append('tmp%n')
        args.append('-d');      args.append(unicode(self._workDir.path()))
        args.append(unicode(self._trackList.audioFile()))
    
        try:
            try:
                proc = subprocess.Popen(args , stdin=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
            except OSError, e:
                msg = self.tr("Splitter program '%1' not found.").arg(qApp.preferences.value("Programs/shntool").toString())
                raise Exception(unicode(msg))
            
            self.sendCueData(proc)
            proc.stdin.close()
            self.parseOut(proc)
            res = proc.wait()
            if res:
                codec = QTextCodec.codecForLocale()              
                for line in proc.stderr:
                    self._errors << codec.toUnicode(line)             
                raise Exception(unicode(self._errors.join('\n')))
                
        except Abort, e:                
            if self._currentTrack > -1:
                self.emit(SIGNAL("progress(int, QString, int)"), self._currentTrack, self.tr('Aborted'), -1)

        except Exception, e: 
            if self._currentTrack > -1:
                self.emit(SIGNAL("progress(int, QString, int)"), self._currentTrack, self.tr('Error'), -1)
              
            self.emit(SIGNAL("error(QString)"), QString(unicode(e)))
            raise 
           

   #****************************************************************
   # Fix shnsplit error: m:ss.ff format can only be used with CD-quality files
   # I don't know last part in frames or in milliseconds?
   #****************************************************************
    def sendCueData(self, proc):
        cdQuality = (self._trackList.audioInfo().item("CD quality") == 'yes')    
    
        cueData = self._trackList.cueData()
        for i in xrange(0, cueData.count()):
            line = cueData[i].trimmed()
            if (not cdQuality) and (line.section(' ', 0, 0).toUpper() == 'INDEX'):
                ff = int(line.section(':', 2, 2))
                line = QString("%1.%2").arg(line.section(':', 0, 1)).arg(ff*10, 3, 10, QChar('0'))
                       
            proc.stdin.write(line.toUtf8())
            proc.stdin.write("\n")
    

   #****************************************************************
   # 
   #****************************************************************
    def parseOut(self, proc):
        codec = QTextCodec.codecForLocale()
        buff = ''
        trackRe = re.compile(r'/tmp(\d{1,4})\.wav\]\s+\(\d{1,2}:\d{2,2}\.\d{1,3}\)\s+:\s*\d+%')

        while True:
            if self._cancel: 
                proc.terminate()
                proc.wait()
                raise Abort()         
        
            c = proc.stderr.read(1)
            if not c: break
            buff += c
            #sys.stdout.write(c)

            #***************************************************
            if (c == "\n"):
                # Errorrs
                if (buff[8:16] == ': error:'):
                    self._errors << codec.toUnicode(buff[17:])

                buff = ''
            #***************************************************
            elif (c == '%'):
                percent = int((buff[-4:-1]))
                r = trackRe.search(buff)
                if (r):
                    self._currentTrack = int(r.groups()[0])-1

                if self._currentTrack > -1:
                    if percent == 100:
                        fileName = QString('%1/tmp%2.wav').arg(self._workDir.path()).arg(self._currentTrack+1, 4, 10, QChar('0'))
                        self.emit(SIGNAL("fileReady(QString, int)"), fileName, self._currentTrack)
                    else:
                        self.emit(SIGNAL("progress(int, QString, int)"), self._currentTrack, self.tr('Extracting'), percent)


   #****************************************************************
   # 
   #****************************************************************
    def stop(self):
        self._cancel = True


   #****************************************************************
   # 
   #****************************************************************
    @staticmethod
    def check():
        if qApp.preferences.value('Programs/shntool').toString().isEmpty():
            raise ProgramNotfound('shntool')



class Encoder(QThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, wavFile, track):
        QThread.__init__(self)
        self._wavFile = wavFile
        self._outFile = track.outFile().filePath()
        self._track = track
        self._cancel = False
        self._proc = None


   #****************************************************************
   #
   #****************************************************************
    def stop(self):
        self._cancel = True
            

   #****************************************************************
   #
   #****************************************************************
    def run(self):
        self._cancel = False
        self.emit(SIGNAL("progress(int, QString, int)"), self._track.index(), self.tr('Encoding'), 0)
        try:
            if self._cancel: raise Abort()
            self.createDir()

            if self._cancel: raise Abort()
            self.encode()

            if self._cancel: raise Abort()
            self.setTags()

            if self._cancel: raise Abort()
            self.deleteWavFile()  

            self.emit(SIGNAL("progress(int, QString, int)"), self._track.index(), self.tr('OK'), -1)

        except Abort, e:
            self.emit(SIGNAL("progress(int, QString, int)"), self._track.index(), self.tr('Aborted'), -1)
        
        except Exception, e:
            self.emit(SIGNAL("progress(int, QString, int)"), self._track.index(), self.tr('Error'), -1)
            self.emit(SIGNAL("error(QString)"), QString(unicode(e)))
            raise
            

   #****************************************************************
   # 
   #****************************************************************
    def createDir(self):
        outDir = self._track.outFile().dir()
        if not outDir.mkpath('.'):
            raise Exception(self.tr("Can't create directory \"%1\".").arg(outDir.path()))


        outFile = QFile(self._track.outFile().filePath())
        if outFile.exists() and not outFile.remove():
            raise Exception(self.tr("Can't delete file:\n%1\n%2").arg(outFile.fileName()).arg(outFile.errorString()))


   #****************************************************************
   # 
   #****************************************************************
    def encode(self):
        qargs = self.getEncoderArgs()
        #print unicode(qargs.join(" "))
        
        args = []
        for s in qargs:
            args.append(unicode(s))


        try:
            self._proc = subprocess.Popen(args , stdin=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        except OSError, e:
            raise Exception(unicode(self.tr('Encoder program "%1" not found.').arg(args[0])))
        
        progress = 0
        
        total = os.path.getsize(unicode(self._wavFile))
        ready = 0 
        buff = ''
        buffSize = int(total / 200)

        f = open(unicode(self._wavFile), 'rb')
        while True:
            
            if self._cancel: 
                self._proc.terminate()
                self._proc.wait()
                raise Abort() 
                
            buff = f.read(buffSize)
            if not buff: break

            ready+=buffSize
            p = int(ready *100 / total)
            if (progress != p):
                progress = p
                self.emit(SIGNAL("progress(int, QString, int)"), self._track.index(), self.tr('Encoding'), progress)
            try:
                self._proc.stdin.write(buff)
            except:
                self.error()
                
        self._proc.stdin.close()
        res = self._proc.wait()
        if res:
            self.error()

    
   #****************************************************************
   #
   #****************************************************************
    def error(self):
        codec = QTextCodec.codecForLocale()
        lines = QStringList()
        for line in self._proc.stderr:
            lines << codec.toUnicode(line)
        
        raise Exception(unicode(lines.join("\n")))


   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        raise TypeError('Abstract method "Encoder.getEncoderArgs" called')
        

   #****************************************************************
   #
   #****************************************************************
    def setTags(self):
        pass


   #****************************************************************
   #
   #****************************************************************
    def deleteWavFile(self):
        f = QFile(self._wavFile)
        if f.exists() and not f.remove():
            raise Exception(self.tr("Can't delete file:\n%1\n%2").arg(f.fileName()).arg(f.errorString()))



class WavEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def encode(self):    
        inFile = QFile(self._wavFile)
        if not inFile.rename(self._outFile):
            raise Exception(self.tr("Can't rename file:\n%1\n%2").arg(self._outFile).arg(inFile.errorString()))


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check():
        pass


class FlacEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/flac").toString()
        args << "--force"       # Force overwriting of output files.

        # Settings .................................................
        args << QString("--compression-level-%1").arg(qApp.preferences.value("Flac/Compression").toString())
        
        # Tags .....................................................
        if (self._track.artist()): args << "--tag" << QString("artist=%1").arg(self._track.artist())
        if (self._track.album()):  args << "--tag" << QString("album=%1").arg(self._track.album())
        if (self._track.genre()):  args << "--tag" << QString("genre=%1").arg(self._track.genre())
        if (self._track.year()):   args << "--tag" << QString("date=%1").arg(self._track.year())
        if (self._track.title()):  args << "--tag" << QString("title=%1").arg(self._track.title())
        if (self._track.discId()): args << "--tag" << QString("discId=%1").arg(self._track.discId())
        args << "--tag" << QString("tracknumber=%1").arg(self._track.trackNum())

        
        args << "-o" << self._outFile,  
        args << "-"
        return args


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check():
        if qApp.preferences.value('Programs/flac').toString().isEmpty():
            raise ProgramNotfound('flac')


class OggEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/oggenc").toString()
        args << "--quiet"
        
        # Quality settings .........................................
        if qApp.preferences.valueAsBool("Ogg/UseQuality"):
            args << "-q" << qApp.preferences.value("Ogg/Quality").toString()
        else:
            val = qApp.preferences.value("Ogg/NormBitrate").toString()
            if val: 
                args << "-b" << val
                
            val = qApp.preferences.value("Ogg/MinBitrate").toString()
            if val: 
                args << "-m" << val

            val = qApp.preferences.value("Ogg/MaxBitrate").toString()
            if val: 
                args << "-M" << val

        # Tags .....................................................
        if (self._track.artist()): args << "--artist"  << self._track.artist()
        if (self._track.album()):  args << "--album"   << self._track.album()
        if (self._track.genre()):  args << "--genre"   << self._track.genre()
        if (self._track.year()):   args << "--date"    << str(self._track.year())
        if (self._track.title()):  args << "--title"   << self._track.title()
        if (self._track.discId()): args << "--comment" << QString("DISCID=%1").arg(self._track.discId())
        args << "--tracknum" << str(self._track.trackNum())
        
        # Files ....................................................
        args << "-o" << self._outFile,  
        args << "-"
        return args
   
   
   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check():
        if qApp.preferences.value('Programs/oggenc').toString().isEmpty():
            raise ProgramNotfound('oggenc')    
            
            
class Mp3Encoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/lame").toString()
        args << "--silent"

        # Settings .................................................
        preset = qApp.preferences.value("Mp3/Preset").toString()
        
        if preset == 'avrMedium': 
            args << "--preset" << "medium"
            
        elif preset == 'avrStandard': 
            args << "--preset" << "standard"   
            
        elif preset == 'avrStandardFast': 
            args << "--preset" << "fast" << "standard"   
            
        elif preset == 'avrExtreme': 
            args << "--preset" << "extreme"   
            
        elif preset == 'avrExtremeFast': 
            args << "--preset" << "fast" << "extreme"   

        elif preset == 'cbrInsane': 
            args << "--preset" << "insane"   
            
        elif preset == 'cbrKbps': 
            args << "--preset" << "cbr" << qApp.preferences.value("Mp3/Bitrate").toString()  

        elif preset == 'abrKbps': 
            args << "--preset" << qApp.preferences.value("Mp3/Bitrate").toString()  

        elif preset == 'vbrQuality': 
            (quality, ok) = qApp.preferences.value("Mp3/Quality").toInt()
            args << "-V" << QString("%1").arg(9 - quality)

        # Tags .....................................................
        args << "--id3v2-only"
        if (self._track.artist()): args << "--ta" << self._track.artist()
        if (self._track.album()):  args << "--tl" << self._track.album()
        if (self._track.genre()):  args << "--tg" << self._track.genre()
        if (self._track.year()):   args << "--ty" << str(self._track.year())
        if (self._track.title()):  args << "--tt" << self._track.title()
        if (self._track.discId()): args << "--tc" << QString("discid=%1").arg(self._track.discId())
        args << "--tn" << str(self._track.trackNum())


        # Files ....................................................
        args << "-"        
        args << self._outFile,  
        
        return args       
        
        
   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check():
        if qApp.preferences.value('Programs/lame').toString().isEmpty():
            raise ProgramNotfound('lame')
            
            
            